"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.BroadcasterLambda = void 0;
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_core_1 = require("@fluidframework/server-services-core");
// Set immediate is not available in all environments, specifically it does not work in a browser.
// Fallback to set timeout in those cases
let taskScheduleFunction;
let clearTaskScheduleTimerFunction;
if (typeof setImmediate === "function") {
    taskScheduleFunction = setImmediate;
    clearTaskScheduleTimerFunction = clearImmediate;
}
else {
    taskScheduleFunction = setTimeout;
    clearTaskScheduleTimerFunction = clearTimeout;
}
/**
 * @internal
 */
class BroadcasterLambda {
    constructor(publisher, context, serviceConfiguration, clientManager) {
        this.publisher = publisher;
        this.context = context;
        this.serviceConfiguration = serviceConfiguration;
        this.clientManager = clientManager;
        this.pending = new Map();
        this.current = new Map();
    }
    async handler(message) {
        const boxcar = (0, server_services_core_1.extractBoxcar)(message);
        for (const baseMessage of boxcar.contents) {
            let topic;
            let event;
            switch (baseMessage.type) {
                case server_services_core_1.SequencedOperationType: {
                    event = "op";
                    const sequencedOperationMessage = baseMessage;
                    topic = `${sequencedOperationMessage.tenantId}/${sequencedOperationMessage.documentId}`;
                    break;
                }
                case server_services_core_1.NackOperationType: {
                    event = "nack";
                    const nackMessage = baseMessage;
                    topic = `client#${nackMessage.clientId}`;
                    break;
                }
                case server_services_core_1.SignalOperationType: {
                    event = "signal";
                    const ticketedSignalMessage = baseMessage;
                    topic = `${ticketedSignalMessage.tenantId}/${ticketedSignalMessage.documentId}`;
                    if (this.clientManager && ticketedSignalMessage.operation) {
                        const signalContent = JSON.parse(ticketedSignalMessage.operation.content);
                        const signalType = typeof signalContent.type === "string" ? signalContent.type : undefined;
                        switch (signalType) {
                            case protocol_definitions_1.MessageType.ClientJoin: {
                                const signalClient = signalContent.content;
                                await this.clientManager.addClient(ticketedSignalMessage.tenantId, ticketedSignalMessage.documentId, signalClient.clientId, signalClient.client, ticketedSignalMessage.operation);
                                break;
                            }
                            case protocol_definitions_1.MessageType.ClientLeave:
                                await this.clientManager.removeClient(ticketedSignalMessage.tenantId, ticketedSignalMessage.documentId, signalContent.content, ticketedSignalMessage.operation);
                                break;
                            default:
                                // ignore unknown types
                                break;
                        }
                    }
                    break;
                }
                default:
                    // ignore unknown types
                    continue;
            }
            const value = baseMessage;
            if (value.type === server_services_core_1.SequencedOperationType &&
                value.operation?.traces &&
                value.operation.traces.length > 0) {
                value.operation.traces.push({
                    action: "start",
                    service: "broadcaster",
                    timestamp: Date.now(),
                });
            }
            if (this.serviceConfiguration.broadcaster.includeEventInMessageBatchName) {
                topic += event;
            }
            let pendingBatch = this.pending.get(topic);
            if (!pendingBatch) {
                pendingBatch = {
                    tenantId: value.tenantId,
                    documentId: value.documentId,
                    event,
                    messages: [value.operation],
                };
                this.pending.set(topic, pendingBatch);
            }
            else {
                pendingBatch.messages.push(value.operation);
            }
        }
        this.pendingOffset = message;
        this.sendPending();
        return undefined;
    }
    close() {
        this.pending.clear();
        this.current.clear();
        this.pendingOffset = undefined;
        if (this.messageSendingTimerId !== undefined) {
            clearTaskScheduleTimerFunction(this.messageSendingTimerId);
            this.messageSendingTimerId = undefined;
        }
    }
    hasPendingWork() {
        return this.pending.size !== 0 || this.current.size !== 0;
    }
    sendPending() {
        if (this.messageSendingTimerId !== undefined) {
            // a send is in progress
            return;
        }
        if (this.pending.size === 0) {
            // no pending work. checkpoint now if we have a pending offset
            if (this.pendingOffset) {
                this.context.checkpoint(this.pendingOffset);
                this.pendingOffset = undefined;
            }
            return;
        }
        // Invoke the next send after a delay to give IO time to create more batches
        this.messageSendingTimerId = taskScheduleFunction(async () => {
            const batchOffset = this.pendingOffset;
            this.current = this.pending;
            this.pending = new Map();
            this.pendingOffset = undefined;
            // Process all the batches + checkpoint
            if (this.publisher.emitBatch) {
                const promises = [];
                for (const [topic, batch] of this.current) {
                    promises.push(this.publisher.emitBatch(topic, batch));
                }
                try {
                    await Promise.all(promises);
                }
                catch (ex) {
                    this.context.error(ex, { restart: true });
                    return;
                }
            }
            else if (this.publisher.emit) {
                const promises = [];
                for (const [topic, batch] of this.current) {
                    promises.push(this.publisher.emit(topic, batch.event, batch.documentId, batch.messages));
                }
                try {
                    await Promise.all(promises);
                }
                catch (ex) {
                    this.context.error(ex, { restart: true });
                    return;
                }
            }
            else {
                for (const [topic, batch] of this.current) {
                    this.publisher.to(topic).emit(batch.event, batch.documentId, batch.messages);
                }
            }
            this.messageSendingTimerId = undefined;
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            this.context.checkpoint(batchOffset);
            this.current.clear();
            this.sendPending();
        });
    }
}
exports.BroadcasterLambda = BroadcasterLambda;
//# sourceMappingURL=lambda.js.map