"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ScriptoriumLambda = void 0;
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const server_services_client_1 = require("@fluidframework/server-services-client");
var ScriptoriumStatus;
(function (ScriptoriumStatus) {
    ScriptoriumStatus["Processing"] = "Processing";
    ScriptoriumStatus["ProcessingComplete"] = "ProcessingComplete";
    ScriptoriumStatus["CheckpointComplete"] = "CheckpointComplete";
    ScriptoriumStatus["ProcessingFailed"] = "ProcessingFailed";
    ScriptoriumStatus["CheckpointFailed"] = "CheckpointFailed";
})(ScriptoriumStatus || (ScriptoriumStatus = {}));
/**
 * @internal
 */
class ScriptoriumLambda {
    constructor(opCollection, context, providerConfig) {
        this.opCollection = opCollection;
        this.context = context;
        this.providerConfig = providerConfig;
        this.pending = new Map();
        this.current = new Map();
        this.clientFacadeRetryEnabled = (0, server_services_core_1.isRetryEnabled)(this.opCollection);
        this.telemetryEnabled = this.providerConfig?.enableTelemetry;
        this.shouldLogInitialSuccessVerbose =
            this.providerConfig?.shouldLogInitialSuccessVerbose ?? false;
        this.maxDbBatchSize = this.providerConfig?.maxDbBatchSize ?? 1000;
        this.restartOnCheckpointFailure = this.providerConfig?.restartOnCheckpointFailure;
    }
    handler(message) {
        const boxcar = (0, server_services_core_1.extractBoxcar)(message);
        for (const baseMessage of boxcar.contents) {
            if (baseMessage.type === server_services_core_1.SequencedOperationType) {
                const value = baseMessage;
                // Remove traces and serialize content before writing to mongo.
                value.operation.traces = [];
                const topic = `${value.tenantId}/${value.documentId}`;
                let pendingMessages = this.pending.get(topic);
                if (!pendingMessages) {
                    pendingMessages = [];
                    this.pending.set(topic, pendingMessages);
                }
                pendingMessages.push(value);
            }
        }
        this.pendingOffset = message;
        if (this.telemetryEnabled && this.pending.size > 0) {
            if (this.pendingMetric === undefined) {
                // create a new metric for processing the current kafka batch
                this.pendingMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ScriptoriumProcessBatch, {
                    timestampQueuedMessage: message.timestamp
                        ? new Date(message.timestamp).toISOString()
                        : null,
                    timestampReadyToProcess: new Date().toISOString(),
                    [server_services_telemetry_1.QueuedMessageProperties.partition]: this.pendingOffset?.partition,
                    [server_services_telemetry_1.QueuedMessageProperties.offsetStart]: this.pendingOffset?.offset,
                    [server_services_telemetry_1.QueuedMessageProperties.offsetEnd]: this.pendingOffset?.offset,
                    [server_services_telemetry_1.CommonProperties.totalBatchSize]: boxcar.contents.length,
                });
            }
            else {
                // previous batch is still waiting to be processed, update properties in the existing metric
                this.pendingMetric.setProperty(server_services_telemetry_1.QueuedMessageProperties.offsetEnd, this.pendingOffset?.offset);
                const currentBatchSize = boxcar.contents.length;
                const previousBatchSize = Number(this.pendingMetric.properties.get(server_services_telemetry_1.CommonProperties.totalBatchSize));
                this.pendingMetric.setProperty(server_services_telemetry_1.CommonProperties.totalBatchSize, previousBatchSize + currentBatchSize);
            }
        }
        this.sendPending();
        return undefined;
    }
    close() {
        this.pending.clear();
        this.current.clear();
        return;
    }
    sendPending() {
        // If there is work currently being sent or we have no pending work return early
        if (this.current.size > 0 || this.pending.size === 0) {
            return;
        }
        let metric;
        if (this.telemetryEnabled && this.pendingMetric) {
            metric = this.pendingMetric;
            this.pendingMetric = undefined;
            metric.setProperty("timestampProcessingStart", new Date().toISOString());
        }
        let status = ScriptoriumStatus.Processing;
        // Swap current and pending
        const temp = this.current;
        this.current = this.pending;
        this.pending = temp;
        const batchOffset = this.pendingOffset;
        const allProcessed = [];
        // Process all the batches + checkpoint
        for (const [, messages] of this.current) {
            if (this.maxDbBatchSize > 0 && messages.length > this.maxDbBatchSize) {
                // cap the max batch size sent to mongo db
                let startIndex = 0;
                while (startIndex < messages.length) {
                    const endIndex = startIndex + this.maxDbBatchSize;
                    const messagesBatch = messages.slice(startIndex, endIndex);
                    startIndex = endIndex;
                    const processP = this.processMongoCore(messagesBatch, metric?.id);
                    allProcessed.push(processP);
                }
            }
            else {
                const processP = this.processMongoCore(messages, metric?.id);
                allProcessed.push(processP);
            }
        }
        Promise.all(allProcessed)
            .then(() => {
            this.current.clear();
            status = ScriptoriumStatus.ProcessingComplete;
            metric?.setProperty("timestampProcessingComplete", new Date().toISOString());
            // checkpoint batch offset
            try {
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                this.context.checkpoint(batchOffset, this.restartOnCheckpointFailure);
                status = ScriptoriumStatus.CheckpointComplete;
                metric?.setProperty("timestampCheckpointComplete", new Date().toISOString());
            }
            catch (error) {
                status = ScriptoriumStatus.CheckpointFailed;
                metric?.setProperty("timestampCheckpointFailed", new Date().toISOString());
                const errorMessage = "Scriptorium failed to checkpoint batch";
                this.logErrorTelemetry(errorMessage, error, status, batchOffset?.offset, metric);
                throw error;
            }
            metric?.setProperty("status", status);
            metric?.success("Scriptorium completed processing and checkpointing of batch");
            // continue with next batch
            this.sendPending();
        })
            .catch((error) => {
            // catches error if any of the promises failed in Promise.all, i.e. any of the ops failed to write to db
            status = ScriptoriumStatus.ProcessingFailed;
            metric?.setProperty("timestampProcessingFailed", new Date().toISOString());
            const errorMessage = "Scriptorium failed to process batch, going to restart";
            this.logErrorTelemetry(errorMessage, error, status, batchOffset?.offset, metric);
            // Restart scriptorium
            this.context.error(error, { restart: true });
        });
    }
    logErrorTelemetry(errorMessage, error, status, batchOffset, metric) {
        if (this.telemetryEnabled && metric) {
            metric.setProperty("status", status);
            metric.error(errorMessage, error);
        }
        else {
            server_services_telemetry_1.Lumberjack.error(errorMessage, { batchOffset, status }, error);
        }
    }
    async processMongoCore(messages, scriptoriumMetricId) {
        return this.insertOp(messages, scriptoriumMetricId);
    }
    async insertOp(messages, scriptoriumMetricId) {
        const dbOps = messages.map((message) => ({
            ...message,
            mongoTimestamp: new Date(message.operation.timestamp),
        }));
        const documentId = messages[0]?.documentId ?? "";
        const tenantId = messages[0]?.tenantId ?? "";
        const sequenceNumbers = messages.map((message) => message.operation.sequenceNumber);
        const sequenceNumberRanges = (0, server_services_client_1.convertSortedNumberArrayToRanges)(sequenceNumbers);
        const insertBatchSize = dbOps.length;
        return (0, server_services_core_1.runWithRetry)(async () => this.opCollection.insertMany(dbOps, false), "insertOpScriptorium", 3 /* maxRetries */, 1000 /* retryAfterMs */, {
            ...(0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId),
            ...{ sequenceNumberRanges, insertBatchSize, scriptoriumMetricId },
        }, (error) => error.code === 11000 ||
            error.message?.toString()?.indexOf("E11000 duplicate key") >= 0, (error) => !this.clientFacadeRetryEnabled /* shouldRetry */, undefined /* calculateIntervalMs */, undefined /* onErrorFn */, this.telemetryEnabled, this.shouldLogInitialSuccessVerbose);
    }
}
exports.ScriptoriumLambda = ScriptoriumLambda;
//# sourceMappingURL=lambda.js.map